from __future__ import absolute_import
from ufl.algorithms import MultiFunction


class SplitIntoAccumulationTerms(MultiFunction):
    """ return a list of tuples of expressions and modified arguments! """

    call = MultiFunction.__call__

    def __call__(self, expr):
        from dune.perftool.ufl.rank import ufl_rank
        self.rank = ufl_rank(expr)

        from dune.perftool.ufl.modified_terminals import ModifiedArgumentExtractor
        self.mae = ModifiedArgumentExtractor()

        from dune.perftool.ufl.transformations.argument_elimination import EliminateArguments
        self.ea = EliminateArguments()

        # Collect the found terms
        self.terms = {}

        # Now, fill the terms dict by applying this multifunction!
        self.call(expr)

        from dune.perftool.ufl.flatoperators import construct_binary_operator
        from ufl.classes import Sum
        return [(construct_binary_operator(t, Sum), a) for a, t in self.terms.items()]

    def expr(self, o):
        # This needs to be a valid accumulation term!
        assert all(len(self.mae(o, i)) == 1 for i in range(self.rank))

        expr, arg = self.ea(o)
        if arg not in self.terms:
            self.terms[arg] = []
        self.terms[arg].append(expr)

    def sum(self, o):
        # Check whether this sums contains too many accumulation terms.
        if not all(len(self.mae(o, i)) == 1 for i in range(self.rank)):
            # This sum is part of a top level sum that separates accumulation terms!
            for op in o.ufl_operands:
                self.call(op)
        else:
            # This is a normal sum, we might treat it as any other expression
            self.expr(o)

    def index_sum(self, o):
        # Check whether this sums contains too many accumulation terms.
        if not all(len(self.mae(o, i)) == 1 for i in range(self.rank)):
            # This sum is part of a top level sum that separates accumulation terms!
            self.call(o.ufl_operands[0])
        else:
            # This is a normal sum, we might treat it as any other expression
            # TODO we need to eliminate topsum indexsum regardless of the thing being valid
            self.expr(o.ufl_operands[0])
            # old code
            #self.expr(o)


def split_into_accumulation_terms(expr):
    return SplitIntoAccumulationTerms()(expr)
